#include "Graphics.h"
#include "CSDL.h"
#include <string.h>
#include "Maths.h"

CGfx gfx;

Color::Color(Uint8 _r, Uint8 _g, Uint8 _b, Uint8 _a)
{
    r = _r;
    g = _g;
    b = _b;
    a = _a;
}

Color::Color(Uint32 _argb)
{
    argb = _argb;
}

CGfx::CGfx()
{
    mPixels2D = NULL;
}

void CGfx::init(const char* title, int w, int h)
{
    sdl.Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);
    mWidth = w;
    mHeight = h;
    mWindow = sdl.CreateWindow(title, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, w, h, SDL_WINDOW_SHOWN);
    mRenderer = sdl.CreateRenderer(mWindow, -1, SDL_RENDERER_ACCELERATED);
}

void CGfx::init2D()
{
    mTexture2D = sdl.CreateTexture(mRenderer, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_STATIC, mWidth, mHeight);
    mPixels2D = new Color[mWidth * mHeight];
    memset(mPixels2D, 0, mWidth * mHeight * sizeof(Uint32));
}

void CGfx::quit()
{
    if (mPixels2D != NULL)
    {
        delete[] mPixels2D;
        sdl.DestroyTexture(mTexture2D);
    }
    sdl.DestroyRenderer(mRenderer);
    sdl.DestroyWindow(mWindow);
    sdl.Quit();
}

void CGfx::clearScreen(Color c)
{
    if (mPixels2D != NULL)
    {
        for (int i = 0; i < mWidth * mHeight; ++i)
            mPixels2D[i] = c;
    }
    else
    {
        sdl.SetRenderDrawColor(mRenderer, c.r, c.g, c.b, c.a);
        sdl.RenderClear(mRenderer);
    }
}

void CGfx::render()
{
    if (mPixels2D != NULL)
    {
        sdl.UpdateTexture(mTexture2D, NULL, mPixels2D, mWidth * sizeof(Uint32));
        sdl.RenderCopy(mRenderer, mTexture2D, NULL, NULL);
    }

    sdl.RenderPresent(mRenderer);
}

void CGfx::rect(int x0, int y0, int x1, int y1, Color c)
{
    for (int x = x0; x <= x1; x++)
    {
        setPixel(x, y0, c);
        setPixel(x, y1, c);
    }

    for (int y = y0 + 1; y < y1; y++)
    {
        setPixel(x0, y, c);
        setPixel(x1, y, c);
    }
}

void CGfx::rectFill(int x0, int y0, int x1, int y1, Color c)
{
    for (int y = y0; y < y1; y++)
        for (int x = x0; x < x1; x++)
            setPixel(x, y, c);
}

void CGfx::line(int x0, int y0, int x1, int y1, Color c)
{
    int dx = x1 - x0;
    int dy = y1 - y0;
    int incX = SGN(dx);
    int incY = SGN(dy);
    dx = ABS(dx);
    dy = ABS(dy);

    if (dy == 0)
    {
        // horizontal line
        for (int x = x0; x != x1 + incX; x += incX)
            setPixel(x, y0, c);
    }
    else if (dx == 0)
    {
        // vertical line
        for (int y = y0; y != y1 + incY; y += incY)
            setPixel(x0, y, c);
    }
    else if (dx >= dy)
    {
        // more horizontal than vertical
        int slope = 2 * dy;
        int error = -dx;
        int errorInc = -2 * dx;
        int y = y0;

        for (int x = x0; x != x1 + incX; x += incX)
        {
            setPixel(x, y, c);
            error += slope;

            if (error >= 0)
            {
                y += incY;
                error += errorInc;
            }
        }
    }
    else
    {
        // more vertical than horizontal
        int slope = 2 * dx;
        int error = -dy;
        int errorInc = -2 * dy;
        int x = x0;

        for (int y = y0; y != y1 + incY; y += incY)
        {
            setPixel(x, y, c);
            error += slope;

            if (error >= 0)
            {
                x += incX;
                error += errorInc;
            }
        }
    }
}

void CGfx::circle(int centerX, int centerY, int radius, Color c)
{
    int x = 0;
    int y = radius;
    int m = 5 - 4 * radius;

    while (x <= y)
    {
        setPixel(centerX + x, centerY + y, c);
        setPixel(centerX + x, centerY - y, c);
        setPixel(centerX - x, centerY + y, c);
        setPixel(centerX - x, centerY - y, c);
        setPixel(centerX + y, centerY + x, c);
        setPixel(centerX + y, centerY - x, c);
        setPixel(centerX - y, centerY + x, c);
        setPixel(centerX - y, centerY - x, c);

        if (m > 0)
        {
            y--;
            m -= 8 * y;
        }

        x++;

        m += 8 * x + 4;
    }
}

void CGfx::circleFill(int centerX, int centerY, int radius, Color c)
{
    int x = 0;
    int y = radius;
    int m = 5 - 4 * radius;

    while (x <= y)
    {
        Color* ptr = &mPixels2D[pixelPos2D(centerX - y, centerY - x)];
        for (int xx = centerX - y; xx <= centerX + y; xx++)
            *ptr++ = c;

        ptr = &mPixels2D[pixelPos2D(centerX - y, centerY + x)];
        for (int xx = centerX - y; xx <= centerX + y; xx++)
            *ptr++ = c;

        if (m > 0)
        {
            ptr = &mPixels2D[pixelPos2D(centerX - x, centerY - y)];
            for (int xx = centerX - x; xx <= centerX + x; xx++)
                *ptr++ = c;

            ptr = &mPixels2D[pixelPos2D(centerX - x, centerY + y)];
            for (int xx = centerX - x; xx <= centerX + x; xx++)
                *ptr++ = c;

            y--;
            m -= 8 * y;
        }

        x++;
        m += 8 * x + 4;
    }
}
